/* 
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations
 * under the License.
 *
 */

package org.apache.muse.util;

import java.lang.reflect.Method;
import java.util.logging.Logger;

import org.w3c.dom.Node;

import org.apache.muse.util.messages.Messages;
import org.apache.muse.util.messages.MessagesFactory;
import org.apache.muse.util.xml.XmlUtils;

/**
 *
 * LoggingUtils is a collection of methods creating and using log files 
 * via the JDK logging API.
 *
 * @author Dan Jemiolo (danj)
 *
 */

public class LoggingUtils
{
    //
    // Used to lookup all exception messages
    //
    private static Messages _MESSAGES = MessagesFactory.get(FileUtils.class);
    
    private static final int BACK_TRACE_LEVEL = 15;

    
    /**
     * 
     * Writes a nicely-formatted message that shows what parameter types 
     * were expected by the Java method and what types were actually used. 
     * The format of the message is:
     * <br><br>
     * <code>
     * Mapping SOAP request to Java method 'myMethodName':<br><br>
     * Parameter types expected by Java method:<br><br>
     * &nbsp;&nbsp;&nbsp;<b>[the expected parameter types]</b><br><br>
     * Parameter types received by the MessageHandler:<br><br>
     * &nbsp;&nbsp;&nbsp;<b>[the actual parameter types]</b><br><br>
     * Messages about this call will appear below...<br>
     * </code>
     *
     * @param log
     *        The Logger that will record the message.
     * 
     * @param method
     *        The Java method being invoked at the time of the error.
     * 
     * @param parameters
     *        The parameter values that were provided by the MessageHandler 
     *        to the Java method.
     *
     */
    public static void logCall(Logger log, Method method, Object[] parameters)
    {
        if (log == null)
            throw new NullPointerException(_MESSAGES.get("NullLogger"));
        
        if (method == null)
            throw new NullPointerException(_MESSAGES.get("NullMethod"));
        
        if (parameters == null)
            throw new NullPointerException(_MESSAGES.get("NullParameters"));
        
        StringBuffer buffer = new StringBuffer();
        buffer.append("[SERVER TRACE] Mapping SOAP request to Java method '");
        buffer.append(method.getName());
        buffer.append("'.\n\n");        
        buffer.append("Parameter types expected by Java method:\n\n");
        buffer.append("\t[");
        
        Class[] paramTypes = method.getParameterTypes();
        
        for (int n = 0; n < paramTypes.length; ++n)
        {
            buffer.append(paramTypes[n].getName());
            
            if (n != paramTypes.length - 1)
                buffer.append(", ");
        }
        
        buffer.append("]\n\n");
        buffer.append("Parameter types received by the MessageHandler:\n\n");
        buffer.append("\t[");
        
        for (int n = 0; n < parameters.length; ++n)
        {
            String type = null;
            
            if (parameters[n] == null)
                type = "null (no type)";
            
            else
                type = parameters[n].getClass().getName();
            
            buffer.append(type);
            
            if (n != parameters.length - 1)
                buffer.append(", ");
        }
        
        buffer.append("]\n\n");
        buffer.append("Messages about this call will appear below...\n");
        
        log.fine(buffer.toString());        
    }
    
    /**
     * 
     * Writes a nicely-formatted message and stack trace to the given log 
     * using the information in the given exception. The format of the 
     * message is:
     * <br><br>
     * <code>
     * There was an error while processing a request:<br><br>
     * <b>the exception message</b><br><br>
     * <b>the stack trace, limited to 15 levels</b><br><br>
     * ------------------------------------------<br>
     * </code>
     *
     * @param log
     *        The Logger that will record the message.
     * 
     * @param error
     *        The exception providing the message and stack trace.
     *
     */
    public static void logError(Logger log, Throwable error)
    {
        if (log == null)
            throw new NullPointerException(_MESSAGES.get("NullLogger"));
        
        if (error == null)
            throw new NullPointerException(_MESSAGES.get("NullError"));
        
        //
        // log format is:
        //
        // There was an error processing a request:
        //
        // The actual error message.
        //     firstMethod()
        //     secondMethod()
        //     thirdMethod()
        //     ...
        //
        // up to BACK_TRACE_LEVEL stack frames
        // 
        StringBuffer buffer = new StringBuffer();
        
        buffer.append("There was an error while processing a request:\n\n");
        
        buffer.append(error.getMessage());
        buffer.append("\n\n\t");
        
        //
        // do this manually, because we only want 15 or so stack 
        // frames - getting the whole engine trace is unhelpful
        //
        StackTraceElement[] stackTrace = error.getStackTrace();
        
        for (int n = 0; n < stackTrace.length && n < BACK_TRACE_LEVEL; ++n)
        {
            buffer.append(stackTrace[n]);
            buffer.append("\n\t");
        }
        
        if (stackTrace.length > BACK_TRACE_LEVEL)
            buffer.append(stackTrace.length - BACK_TRACE_LEVEL + " more...");
        
        //
        // separator between log messages
        //
        buffer.append("\n------------------------------------------\n");
        
        log.info(buffer.toString());
    }
    
    public static void logMessage(Logger log, Node soap, boolean incoming)
    {
        if (log == null)
            throw new NullPointerException(_MESSAGES.get("NullLogger"));

        if (soap == null)
            throw new NullPointerException(_MESSAGES.get("NullMessageElement"));
        
        StringBuffer buffer = new StringBuffer();
        
        buffer.append("[SERVER TRACE] SOAP envelope contents (");
        buffer.append(incoming ? "incoming" : "outgoing");
        buffer.append("):\n\n");
        buffer.append(XmlUtils.toString(soap, false));
        buffer.append('\n'); 
        
        log.fine(buffer.toString());
    }
}
